require 'slither'

VISIBLE_DISTANCE = 100
REPULSE_DISTANCE = 10
TURNING_DISTANCE = 120

class 'Pos' {
	x = 0,
	y = 0,
	z = 0,
	move = function(self, v)
		self.x = self.x + (v.x or 0)
		self.y = self.y + (v.y or 0)
		self.z = self.z + (v.z or 0)
	end,
	distance = function(self, other)
		return math.sqrt((self.x - other.x)^2 + (self.y - other.y)^2 + (self.z - other.z)^2)
	end,
	__add__ = function(self, other)
		local r = Pos()
		r.x = self.x + other.x
		r.y = self.y + other.y
		r.z = self.z + other.z
		return r
	end,
	__sub__ = function(self, other)
		local r = Pos()
		r.x = self.x - other.x
		r.y = self.y - other.y
		r.z = self.z - other.z
		return r
	end,
	__div__ = function(self, nr)
		local r = Pos()
		r.x = self.x / nr
		r.y = self.y / nr
		r.z = self.z / nr
		return r
	end,
}

class 'Vel' ('Pos') {
	x = 0,
	y = 0,
	z = 0,
	v1 = function(self, b, blist)
		local N = 0
		local total = Pos()
		for _,boid in ipairs(blist) do
			if --[[boid ~= b and]] boid.pos:distance(b.pos) < VISIBLE_DISTANCE then
				N = N + 1
				total:move(boid.pos)
			end
		end
		return (total / N - b.pos) / 100
	end,
	v2 = function(self, b, blist)
		local total = Pos()
		for _,boid in ipairs(blist) do
			if --[[boid ~= b and]] boid.pos:distance(b.pos) < REPULSE_DISTANCE then
				total:move(b.pos - boid.pos)
			end
		end
		return total / 8
	end,
	v3 = function(self, b, blist)
		local N = 0
		local total = Vel()
		for _,boid in ipairs(blist) do
			if --[[boid ~= b and]] boid.pos:distance(b.pos) < TURNING_DISTANCE then
				N = N + 1
				total:move(boid.vel)
			end
		end
		return (total / N - b.vel) / 8
	end,
	lock = function(self, b, blist)
		local total = Vel()
		if b.pos.x < -390 then
			total:move{x = 10}
		elseif b.pos.x > 390 then
			total:move{x = -10}
		end
		if b.pos.y < -390 then
			total:move{y = 10}
		elseif b.pos.y > 390 then
			total:move{y = -10}
		end
		if b.pos.z < -390 then
			total:move{z = 10}
		elseif b.pos.z > 390 then
			total:move{z = -10}
		end
		return total
	end,
	adapt = function(self, b, blist)
		self:move(self:v1(b, blist))
		self:move(self:v2(b, blist))
		self:move(self:v3(b, blist))
		self:move(self:lock(b, blist))
	end
}

class 'Boid' {
	__init__ = function(self)
		self.pos = Pos()
		self.vel = Vel()
	end,
	accel = function(self, blist)
		self.vel:adapt(self, blist)
	end,
	move = function(self)
		self.pos:move(self.vel)
		if self.pos.x < -400 then
			self.pos.x = -400
		elseif self.pos.x > 400 then
			self.pos.x = 400
		end
		if self.pos.y < -400 then
			self.pos.y = -400
		elseif self.pos.y > 400 then
			self.pos.y = 400
		end
		if self.pos.z < -400 then
			self.pos.z = -400
		elseif self.pos.z > 400 then
			self.pos.z = 400
		end
	end
}

class 'Simulation' {
	__init__ = function(self, n, maxx)
		self.boids = {}
		for i = 1, n do
			self.boids[i] = Boid()
			self.boids[i].pos.x = math.random(maxx) - maxx / 2
			self.boids[i].pos.y = math.random(maxx) - maxx / 2
			self.boids[i].pos.z = math.random(maxx) - maxx / 2
		end
	end,
	step = function(self)
		for _,b in ipairs(self.boids) do
			b:accel(self.boids)
		end
		for _,b in ipairs(self.boids) do
			b:move()
		end
	end
}

